// 
// Copyright © 2019 Samuel Venable
// Copyright © 2014 Seth N. Hetu
// 
// This file is a part of the ENIGMA Development Environment.
// 
// ENIGMA is free software: you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation, version 3 of the license or any later version.
// 
// This application and its source code is distributed AS-IS, WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
// 
// You should have received a copy of the GNU General Public License along
// with this code. If not, see <http://www.gnu.org/licenses/>
//

#include "IniFileIndex.h"

#include <cstdlib>
#include <algorithm>
#include <sstream>
#include <fstream>

using std::map;
using std::string;
using std::vector;
using std::getline;
using std::ifstream;
using std::ofstream;
using std::stringstream;
using std::ostringstream;

namespace {

string ltrim(const string& str) {
  size_t start = str.find_first_not_of(" \t\r");
  if (start != string::npos) {
    return str.substr(start);
  }
  return str;
}

string rtrim(const string& str) {
  size_t last = str.find_last_not_of(" \t\r");
  if (last != string::npos) {
    return str.substr(0, last+1);
  }
  return str;
}

string trim(const string& str) {
  return rtrim(ltrim(str));
}

string optional_quote(const string& str, const char commentChar) {
  if (!str.empty() && (str[0]==' ' || str[0]=='\t' || str[str.size()-1]==' ' || str[str.size()-1]=='\t' || str.find(commentChar)!=string::npos)) {
    return string("\"") + str + "\"";
  }
  return str;
}

}


void ini::IniFileIndex::load(std::istream& input, char commentChar) {
  this->commentChar = commentChar;

  string secName;

  stringstream comment;

  string line;
  string origLine;
  while(getline(input, origLine)) {
    line = trim(origLine);

    if (line.empty() || line[0]==commentChar) {
      comment <<origLine <<"\n";
    } else if (line[0]=='[') {
      size_t br = line.find(']', 1);
      if (br != string::npos) {
        secName = line.substr(1, br-1);

        if (secName.find('[')==string::npos) {
          if (sections.count(secName)==0) {
            sections[secName].prefix = comment.str();
            secNameOrder.push_back(secName);
          }
        }
      }
      comment.str("");
    } else {
      size_t eq = line.find('=');
      if (eq != string::npos) {
        string key = trim(line.substr(0,eq));
        string value;

        string rhs = trim(line.substr(eq+1));
        if (!rhs.empty() && rhs[0]=='"') {
          size_t qt = rhs.find('"', 1);
          if (qt==string::npos) {
            rhs = "";
          } else {
            value = rhs.substr(1,qt-1);
            rhs = trim(rhs.substr(qt+1));
          }
        } else {
          size_t ct = rhs.find(commentChar, 1);
          if (ct==string::npos) {
            value = rhs;
            rhs = "";
          } else {
            value = trim(rhs.substr(0,ct));
            rhs = trim(rhs.substr(ct));
          }
        }

        if (rhs.empty() || rhs[0]==commentChar) {
          if (!value.empty()) {
            if (sections[secName].props.count(key)==0) {
              sections[secName].props[key].prefix = comment.str();
              sections[secName].props[key].value = value;
              sections[secName].propKeyOrder.push_back(key);
            }
          }
        }

        comment.str("");

        if (rhs[0]==commentChar) {
          sections[secName].props[key].postfix = trim(rhs);
        }
      }
    }
  }

  postComment = comment.str();
}


void ini::IniFileIndex::clear() {
  sections.clear();
  secNameOrder.clear();
  postComment.clear();
}

bool ini::IniFileIndex::keyExists(const string& section, const string& key) {
  map<string, IniFileSection>::const_iterator secIt = sections.find(section);
  if (secIt==sections.end()) { return false; }
  map<string, IniValue>::const_iterator propIt = secIt->second.props.find(key);
  if (propIt==secIt->second.props.end()) { return false; }
  return true;
}

bool ini::IniFileIndex::sectionExists(const string& section) {
  map<string, IniFileSection>::const_iterator secIt = sections.find(section);
  if (secIt==sections.end()) { return false; }
  return true;
}

string ini::IniFileIndex::read(const string& section, const string& key, string def) const {
  map<string, IniFileSection>::const_iterator secIt = sections.find(section);
  if (secIt==sections.end()) { return def; }
  map<string, IniValue>::const_iterator propIt = secIt->second.props.find(key);
  if (propIt==secIt->second.props.end()) { return def; }
  return propIt->second.value;
}

float ini::IniFileIndex::read(const string& section, const string& key, float def) const {
  map<string, IniFileSection>::const_iterator secIt = sections.find(section);
  if (secIt==sections.end()) { return def; }
  map<string, IniValue>::const_iterator propIt = secIt->second.props.find(key);
  if (propIt==secIt->second.props.end()) { return def; }

  return (float)atof(propIt->second.value.c_str());
}

void ini::IniFileIndex::write(const string& section, const string& key, const string& value) {
  map<string, IniFileSection>::const_iterator secIt = sections.find(section);
  if (secIt==sections.end()) {
    secNameOrder.push_back(section);
  }
  map<string, IniValue>::const_iterator propIt = sections[section].props.find(key);
  if (propIt==sections[section].props.end()) {
    sections[section].propKeyOrder.push_back(key);
  }
  sections[section].props[key].value = value;
}

void ini::IniFileIndex::write(const string& section, const string& key, float value) {
  map<string, IniFileSection>::const_iterator secIt = sections.find(section);
  if (secIt==sections.end()) {
    secNameOrder.push_back(section);
  }
  map<string, IniValue>::const_iterator propIt = sections[section].props.find(key);
  if (propIt==sections[section].props.end()) {
    sections[section].propKeyOrder.push_back(key);
  }

  ostringstream valF;
  valF<<value;
  sections[section].props[key].value = valF.str();
}

void ini::IniFileIndex::delKey(const string& section, const string& key) {
  map<string, IniFileSection>::iterator secIt = sections.find(section);
  if (secIt==sections.end()) {
    return;
  }

  map<string, IniValue>::iterator propIt=secIt->second.props.find(key);
  if (propIt==secIt->second.props.end()) {
    return;
  }

  vector<string>::iterator remIt = std::find(secIt->second.propKeyOrder.begin(), secIt->second.propKeyOrder.end(), key);
  if (remIt != secIt->second.propKeyOrder.end()) {
    secIt->second.propKeyOrder.erase(remIt);
  }

  secIt->second.props.erase(propIt);

  if (secIt->second.props.empty()) {
    delSection(section);
  }
}


void ini::IniFileIndex::delSection(const string& section) {
  map<string, IniFileSection>::iterator secIt = sections.find(section);
  if (secIt==sections.end()) {
    return;
  }

  vector<string>::iterator remIt = std::find(secNameOrder.begin(), secNameOrder.end(), section);
  if (remIt != secNameOrder.end()) {
    secNameOrder.erase(remIt);
  }

  sections.erase(secIt);
}


bool ini::IniFileIndex::saveToFile(const string& fname) const {
  if (sections.empty() && postComment.empty()) {
    ifstream file(fname.c_str());
    if (!file.good()) {
      return true;
    }
  }

  ofstream out(fname.c_str());
  if (!out.good()) {
    return false;
  }
  
  out <<this->toString();
  return true;
}

string ini::IniFileIndex::toString() const {
  stringstream res;
  for (vector<string>::const_iterator secIt=secNameOrder.begin(); secIt!=secNameOrder.end(); secIt++) {
    const IniFileSection& sec = sections.find(*secIt)->second;
    res <<sec.prefix;
    res <<"[" <<(*secIt) <<"]\n";
    for (vector<string>::const_iterator keyIt=sec.propKeyOrder.begin(); keyIt!=sec.propKeyOrder.end(); keyIt++) {
      const IniValue& val = sec.props.find(*keyIt)->second;
      res <<val.prefix;
      res <<*keyIt <<" = " <<optional_quote(val.value, commentChar) <<(val.postfix.empty()?"":string(" ")+val.postfix) <<"\n";
    }
  }
  res <<postComment;
  return res.str();
}
